"""LightningFlower Model"""

import pytorch_lightning as pl


class LightningFlowerModel:
    def __init__(self, model):
        assert isinstance(model, pl.LightningModule)
        self.model = model
        # model parameter that are not sent to the server, e.g. frozen weights
        self.fixed_model_parameters = self.__get_fixed_model_params()

    def __get_fixed_model_params(self):
        param_list = []
        for name, param in self.model.named_parameters():
            if not param.requires_grad:
                param_list.append(name)
        return param_list




"""LightningFlower Data Module"""

import os
import torch.utils
import numpy as np
from torch.utils.data import DataLoader
from lightningflower.config import LightningFlowerDefaults, DatasetConfig


class CustomTensorDataset(torch.utils.data.Dataset):
    """TensorDataset with support of transforms."""

    def __init__(self, tensors, transform=None):
        assert all(tensors[0].size(0) == tensor.size(0) for tensor in tensors)
        self.tensors = tensors
        self.transform = transform

    def __getitem__(self, index):
        x = self.tensors[0][index]
        y = self.tensors[1][index]
        if self.transform:
            x = self.transform(x.numpy().astype(np.uint8))
        return x, y

    def __len__(self):
        return self.tensors[0].size(0)


class CustomNumpyDataset(torch.utils.data.Dataset):
    """NumpyDataset with support of transforms."""

    def __init__(self, path, train):
        self.path = path
        if train:
            self.tensors = (
                torch.from_numpy(np.load(os.path.join(path, "X_train.npy"))),
                torch.from_numpy(np.load(os.path.join(path, "y_train.npy")))
            )

        else:
            self.tensors = (
                torch.from_numpy(np.load(os.path.join(path, "X_test.npy"))),
                torch.from_numpy(np.load(os.path.join(path, "y_test.npy")))
            )
        self.data = self.tensors[0].squeeze().float()
        self.targets = self.tensors[-1].float()

    def __getitem__(self, index):
        x = self.tensors[0][index]
        y = self.tensors[1][index]
        return x, y

    def __len__(self):
        return self.tensors[0].size(0)


class LightningFlowerData:
    @staticmethod
    def add_data_specific_args(parent_parser):
        parser = parent_parser.add_argument_group("LightningFlowerData")
        parser.add_argument("--dataset_path", type=str, default=LightningFlowerDefaults.DATASET_FOLDER)
        parser.add_argument("--batch_size_train", type=int, default=LightningFlowerDefaults.BATCH_SIZE_TRAIN)
        parser.add_argument("--batch_size_test", type=int, default=LightningFlowerDefaults.BATCH_SIZE_TEST)
        parser.add_argument("--dataset_config", type=int, default=LightningFlowerDefaults.DATASET_CONFIG,
                            help="0 for NON-IID, 1 for IID, 2 for SAMPLED")
        parser.add_argument("--num_workers", type=int, default=LightningFlowerDefaults.NUM_WORKERS)
        return parent_parser

    @staticmethod
    def get_loader(dataset_train, dataset_test, args):
        train_loader = None
        test_loader = None
        if args.dataset_config is DatasetConfig.NON_IID:
            # Todo Implement
            pass
        elif args.dataset_config is DatasetConfig.IID:
            train_loader = torch.utils.data.DataLoader(dataset=dataset_train,
                                                       batch_size=args.batch_size_train,
                                                       num_workers=args.num_workers,
                                                       drop_last=True)
            test_loader = torch.utils.data.DataLoader(dataset=dataset_test,
                                                      batch_size=args.batch_size_test,
                                                      num_workers=args.num_workers,
                                                      drop_last=True)
        elif args.dataset_config is DatasetConfig.SAMPLED:
            # Todo Implement
            pass
        return train_loader, test_loader

    @staticmethod
    def split_datasets(training_dataset, num_clients, num_shards, iid, transform=None):
        """Split the whole dataset in IID or non-IID manner for distributing to clients."""

        # unsqueeze channel dimension for grayscale image datasets
        if training_dataset.data.ndim == 3:  # convert to NxHxW -> NxHxWx1
            training_dataset.data.unsqueeze_(3)
        num_categories = np.unique(training_dataset.targets).shape[0]

        if "ndarray" not in str(type(training_dataset.data)):
            training_dataset.data = np.asarray(training_dataset.data)
        if "list" not in str(type(training_dataset.targets)):
            training_dataset.targets = training_dataset.targets.tolist()

        # split dataset according to iid flag
        if iid:
            # shuffle data
            shuffled_indices = torch.randperm(len(training_dataset))
            training_inputs = training_dataset.data[shuffled_indices]
            training_labels = torch.Tensor(training_dataset.targets)[shuffled_indices]

            # partition data into num_clients
            split_size = len(training_dataset) // num_clients
            split_datasets = list(
                zip(
                    torch.split(torch.Tensor(training_inputs), split_size),
                    torch.split(torch.Tensor(training_labels), split_size)
                )
            )

            # finalize bunches of local datasets
            local_datasets = [
                CustomTensorDataset(local_dataset, transform=transform)
                for local_dataset in split_datasets
            ]
        else:
            # sort data by labels
            sorted_indices = torch.argsort(torch.Tensor(training_dataset.targets))
            training_inputs = training_dataset.data[sorted_indices]
            training_labels = torch.Tensor(training_dataset.targets)[sorted_indices]

            # partition data into shards first
            shard_size = len(training_dataset) // num_shards  # 300
            shard_inputs = list(torch.split(torch.Tensor(training_inputs), shard_size))
            shard_labels = list(torch.split(torch.Tensor(training_labels), shard_size))

            # sort the list to conveniently assign samples to each clients from at least two classes
            shard_inputs_sorted, shard_labels_sorted = [], []
            for i in range(num_shards // num_categories):
                for j in range(0, ((num_shards // num_categories) * num_categories),
                               (num_shards // num_categories)):
                    shard_inputs_sorted.append(shard_inputs[i + j])
                    shard_labels_sorted.append(shard_labels[i + j])

            # finalize local datasets by assigning shards to each client
            shards_per_clients = num_shards // num_clients
            local_datasets = [
                CustomTensorDataset(
                    (
                        torch.cat(shard_inputs_sorted[i:i + shards_per_clients]),
                        torch.cat(shard_labels_sorted[i:i + shards_per_clients]).long()
                    ),
                    transform=transform
                )
                for i in range(0, len(shard_inputs_sorted), shards_per_clients)
                ]
        return local_datasets



"""LightningFlower Client"""

import flwr as fl
from flwr.common import EvaluateIns, EvaluateRes, FitRes
import pytorch_lightning as pl
import torch
import timeit
from collections import OrderedDict
from lightningflower.utility import id_generator
from lightningflower.config import LightningFlowerDefaults
from lightningflower.model import LightningFlowerModel


class LightningFlowerClient(fl.client.Client):
    @staticmethod
    def add_client_specific_args(parent_parser):
        parser = parent_parser.add_argument_group("LightningFlowerClient")
        parser.add_argument("--host_address", type=str, default=LightningFlowerDefaults.HOST_ADDRESS)
        parser.add_argument("--max_msg_size", type=int, default=LightningFlowerDefaults.GRPC_MAX_MSG_LENGTH)
        return parent_parser

    @staticmethod
    def update_state_dict(current_model_dict, parameters):
        params_dict = zip(current_model_dict.keys(), parameters)
        # in case of trouble with batch norm try torch.from_numpy(v)
        loaded_state_dict = OrderedDict({k: torch.tensor(v) for k, v in params_dict})
        # allow mismatching shapes
        new_state_dict = {k: v if v.size() == current_model_dict[k].size() else current_model_dict[k] for k, v in
                          zip(current_model_dict.keys(), loaded_state_dict.values())}
        return new_state_dict

    def __init__(self,
                 model,
                 trainer_args,
                 c_id=id_generator(),
                 train_loader=None,
                 test_loader=None):
        # make sure that the model is a lightningflowermodel
        assert isinstance(model, LightningFlowerModel)
        # client id
        self.c_id = c_id
        self.localModel = model
        # optional training and test loader
        self.trainLoader = train_loader
        self.testLoader = test_loader
        # trainer configuration
        self.trainer_config = trainer_args

    def get_weights(self):
        """Get model weights as a list of NumPy ndarrays."""
        ret_val = [val.cpu().numpy() for key, val in self.localModel.model.state_dict().items()]
        return ret_val

    def get_trainable_weights(self):
        """Get only trainable model weights as a list of NumPy ndarrays."""
        ret_val = [val.cpu().numpy() for key, val in self.localModel.model.state_dict().items()
                   if key not in self.localModel.fixed_model_parameters]
        return ret_val

    def get_parameters(self):
        """Return the current local model parameters.

        Returns
        -------
        ParametersRes
            The current local model parameters.
        """
        print(f"Client {self.c_id}: get_parameters")
        weights: fl.common.Weights = self.get_trainable_weights()
        parameters = fl.common.weights_to_parameters(weights)
        return fl.common.ParametersRes(parameters=parameters)

    def set_parameters(self, parameters):
        """Set the current local model parameters

        Parameters
        ----------
        parameters : model parameters

        """
        print(f"Client {self.c_id}: set_parameters")
        # we need to adapt for pot. new parameters and shape mismatching, may raise issues later
        # see https://github.com/pytorch/pytorch/issues/40859
        new_state_dict = self.update_state_dict(self.localModel.model.state_dict(), parameters)
        self.localModel.model.load_state_dict(new_state_dict, strict=False)

    def fit(self, ins):
        """Refine the provided weights using the locally held dataset.

        Parameters
        ----------
        ins : FitIns
            The training instructions containing (global) model parameters
            received from the server and a dictionary of configuration values
            used to customize the local training process.

        Returns
        -------
        FitRes
            The training result containing updated parameters and other details
            such as the number of local training examples used for training.
        """
        print(f"Client {self.c_id}: fit")
        weights: fl.common.Weights = fl.common.parameters_to_weights(ins.parameters)
        # read configuration
        config = ins.config
        # begin time measurement
        fit_begin = timeit.default_timer()
        # update local client model parameters
        self.set_parameters(weights)
        # training procedure
        trainer = pl.Trainer.from_argparse_args(self.trainer_config)
        trainer.fit(model=self.localModel.model, train_dataloader=self.trainLoader)
        # calculate nr. of examples used by Trainer for train
        num_train_examples = self.trainLoader.batch_size * trainer.num_training_batches
        # log train data
        #printf("Train: num_train_examples=" + str(num_train_examples))

        # return updated model parameters
        weights_prime: fl.common.Weights = self.get_trainable_weights()
        params_prime = fl.common.weights_to_parameters(weights_prime)
        metrics = {"duration": timeit.default_timer() - fit_begin}
        return FitRes(parameters=params_prime,
                      num_examples=num_train_examples,
                      metrics=metrics)

    def evaluate(self, ins):
        """Evaluate the provided weights using the locally held dataset.

        Parameters
        ----------
        ins : EvaluateIns
            The evaluation instructions containing (global) model parameters
            received from the server and a dictionary of configuration values
            used to customize the local evaluation process.

        Returns
        -------
        EvaluateRes
            The evaluation result containing the loss on the local dataset and
            other details such as the number of local data examples used for
            evaluation.
        """
        print(f"Client {self.c_id}: evaluate")
        weights: fl.common.Weights = fl.common.parameters_to_weights(ins.parameters)
        # update local client model parameters
        self.set_parameters(weights)
        # evaluation procedure
        trainer = pl.Trainer.from_argparse_args(self.trainer_config)
        test_result = trainer.test(self.localModel.model, self.testLoader)
        # obtain result of first train_loader
        train_loader_0_result = test_result[0]
        test_loss = train_loader_0_result["test_loss"]
        accuracy = train_loader_0_result["test_acc"]
        # calculate nr. of examples used by Trainer for test
        num_test_examples = self.testLoader.batch_size * trainer.num_test_batches[0]
        # log test result
        #printf("Evaluation: test_loss=" + str(test_loss) + ", num_test_examples=" + str(num_test_examples) + ", accuracy=" + str(accuracy))
        metrics = {"accuracy": accuracy}
        return EvaluateRes(loss=test_loss,
                           num_examples=num_test_examples,
                           metrics=metrics)



from setuptools import setup

required_packages = [
    "torch", "flwr", "pytorch-lightning", "numpy"
]

setup(
    name="lightningflower",
    version="0.1.1",
    description="Flower based Pytorch lightning framework",
    url="https://github.com/ManuelRoeder/lightningflower",
    author="Manuel Roeder",
    author_email="manuel.roeder@web.de",
    license="MIT",
    packages=["lightningflower"],
    install_requires=required_packages,
    python_requires='>=3.8.12',
    package_data={"": ["README.md", "LICENSE"]},
)



"""LightningFlower Strategy"""

import os
import flwr as fl
import numpy as np
import pytorch_lightning.utilities.argparse as arg_parser
from argparse import ArgumentParser, Namespace
from flwr.server.strategy import FedAvg
from lightningflower.utility import printf, load_params, boolean_string, create_dir
from lightningflower.config import LightningFlowerDefaults
from typing import Any, Union


class LightningFlowerBaseStrategy:
    """Base class for lightning flower strategies.

    Allows saving and loading of initial model weights,
    takes care of argument parsing.
    """

    def __init__(self, init_model, save_model, model_path):
        # create save/load directory
        create_dir(model_path)
        self.save_model = save_model
        self.init_model = init_model
        self.full_weights_path = os.path.join(model_path, LightningFlowerDefaults.WEIGHTS_FILENAME)
        if self.init_model:
            self.initial_parameters = load_params(self.full_weights_path)

    @classmethod
    def from_argparse_args(cls: Any, args: Union[Namespace, ArgumentParser], **kwargs) -> Any:
        """ Creates a LightingFlowerStrategy object from argument parser

        :param args: Arguments
        :param kwargs: Keyword Arguments
        :return: Configured LightningFlowerStrategy object
        """
        return arg_parser.from_argparse_args(cls, args, **kwargs)

    @staticmethod
    def add_strategy_specific_args(parent_parser):
        """ Add dedicated argument group for strategy implementation

        Example:
            parser = parent_parser.add_argument_group("MyStrategy")

            parser.add_argument('--argument', default=0, type=int, help="This is an argument")

            return parser

        :param parent_parser: The parent parser to add argument group
        :return: The updated parent parser
        """
        # LightningFlowerBaseStrategy specific arguments
        parser = parent_parser.add_argument_group("LightningFlowerBaseStrategy")
        parser.add_argument('--init_model', default=False, type=boolean_string,
                            help="Load pretrained global weights if True")
        parser.add_argument("--save_model", default=False, type=boolean_string,
                            help="Save global model weights after each evaluation round")
        parser.add_argument("--model_path", default=LightningFlowerDefaults.WEIGHTS_FOLDER, type=str,
                            help="Path to load/save weights to")
        return parent_parser

    def save_weights(self, weights):
        """ Save the model weights if required and valid

        :param weights:
        :return:
        """
        if weights is not None and self.save_model:
            # Save aggregated_weights
            printf("Saving model weights")
            np.savez(self.full_weights_path, *weights)


class LightningFlowerFedAvgStrategy(LightningFlowerBaseStrategy, fl.server.strategy.FedAvg):
    @staticmethod
    def add_strategy_specific_args(parent_parser):
        # add base strategy related arguments
        parser = LightningFlowerBaseStrategy.add_strategy_specific_args(parent_parser)
        # FedAvg specific arguments
        parser.add_argument("--fraction_fit", type=float, default=0.1)
        parser.add_argument("--fraction_eval", type=float, default=0.1)
        parser.add_argument("--min_fit_clients", type=int, default=2)
        parser.add_argument("--min_eval_clients", type=int, default=2)
        parser.add_argument("--min_available_clients", type=int, default=2)
        parser.add_argument("--accept_failures", type=boolean_string, default=True)
        return parent_parser

    def __init__(self, fraction_fit,
                       fraction_eval,
                       min_fit_clients,
                       min_eval_clients,
                       min_available_clients,
                       accept_failures,
                       init_model,
                       save_model,
                       model_path):
        fl.server.strategy.FedAvg.__init__(self,
                                           fraction_fit=fraction_fit,
                                           fraction_eval=fraction_eval,
                                           min_fit_clients=min_fit_clients,
                                           min_eval_clients=min_eval_clients,
                                           min_available_clients=min_available_clients,
                                           accept_failures=accept_failures)

        LightningFlowerBaseStrategy.__init__(self,
                                             init_model=init_model,
                                             save_model=save_model,
                                             model_path=model_path)

    def aggregate_fit(self, rnd, results, failures):
        # dispatch weight aggregation from FedAvg strategy
        aggregated_weights = fl.server.strategy.FedAvg.aggregate_fit(self, rnd, results, failures)
        # save weights if required
        LightningFlowerBaseStrategy.save_weights(self, aggregated_weights)
        return aggregated_weights
